# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    return_type_single_tab.py                          :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: charles <me@cacharle.xyz>                  +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/01/09 10:28:30 by charles           #+#    #+#              #
#    Updated: 2022/01/09 10:28:30 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #


import re

from c_formatter_42.formatters import helper


def return_type_single_tab(content: str) -> str:
    lines = content.split("\n")
    aligned = []
    align_regex = (
        r"^(?P<prefix>{type})\s+"
        r"(?P<suffix>{name}\(.*\)?[^;])$"
    )
    align_regex = align_regex.format(
        type=helper.REGEX_TYPE,
        name=helper.REGEX_NAME,
        decl=helper.REGEX_DECL_NAME
    )
    matches = [re.match(align_regex, line) for line in lines]
    aligned = [(i, match.group("prefix"), match.group("suffix"))
               for i, match in enumerate(matches)
               if match is not None and
               match.group("prefix") not in ["struct", "union", "enum"]]
    for i, prefix, suffix in aligned:
        lines[i] = f"{prefix}\t{suffix}"
    return "\n".join(lines)



#!/usr/bin/env python3

# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    main.py                                            :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/04 09:53:21 by cacharle          #+#    #+#              #
#    Updated: 2020/10/04 09:53:21 by cacharle         ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import sys
import argparse

from c_formatter_42.run import run_all


def main():
    arg_parser = argparse.ArgumentParser(
        prog="c_formatter_42",
        description="Format C source according to the norm",
        formatter_class=argparse.RawTextHelpFormatter
    )
    arg_parser.add_argument(
        "-c", "--confirm",
        action="store_true",
        help="Ask confirmation before overwritting any file"
    )
    arg_parser.add_argument(
        "filepaths",
        metavar="FILE",
        nargs="*",
        help="File to format inplace, if no file is provided read STDIN"
    )
    args = arg_parser.parse_args()

    if len(args.filepaths) == 0:
        content = sys.stdin.read()
        print(run_all(content), end="")
    else:
        for filepath in args.filepaths:
            try:
                with open(filepath, "r") as file:
                    content = file.read()
                if args.confirm:
                    result = input("Are you sure you want to overwrite {}?[y/N]"
                                   .format(filepath))
                    if result != "y":
                        continue
                print("Writting to {}".format(filepath))
                with open(filepath, "w") as file:
                    file.write(run_all(content))
            except OSError as e:
                print("Error: {}: {}".format(e.filename, e.strerror))


if __name__ == "__main__":
    main()



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    run.py                                             :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/09 18:15:09 by cacharle          #+#    #+#              #
#    Updated: 2021/02/11 20:12:15 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #

from c_formatter_42.formatters.clang_format import clang_format
from c_formatter_42.formatters.hoist import hoist
from c_formatter_42.formatters.align import align
from c_formatter_42.formatters.preprocessor_directive import preprocessor_directive
from c_formatter_42.formatters.return_type_single_tab import return_type_single_tab
from c_formatter_42.formatters.misc import (
    parenthesize_return,
    space_before_semi_colon,
    remove_multiline_condition_space,
    insert_void,
)
from c_formatter_42.formatters.line_breaker import line_breaker


def run_all(content: str) -> str:
    """Run all formatters"""
    content = clang_format(content)
    content = preprocessor_directive(content)
    content = remove_multiline_condition_space(content)
    content = parenthesize_return(content)
    content = space_before_semi_colon(content)
    content = hoist(content)
    content = align(content)
    content = return_type_single_tab(content)
    content = insert_void(content)
    content = line_breaker(content)
    return content



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    misc.py                                            :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: charles <me@cacharle.xyz>                  +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2021/02/07 14:39:26 by charles           #+#    #+#              #
#    Updated: 2021/02/07 21:16:03 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import re


def parenthesize_return(content: str) -> str:
    return re.sub(
        r"return\s+(?!;)(?!\(.*\);)(?P<value>\(?.*?)\s*;",
        lambda match: "return ({});".format(match.group("value").strip()),
        content,
        re.DOTALL,
    )


def space_before_semi_colon(content: str) -> str:
    return re.sub(
        r"(?P<keyword>return|break|continue);",
        lambda match: match.group("keyword") + " ;",
        content,
    )


def remove_multiline_condition_space(content: str) -> str:
    return re.sub(
        r"(?P<tabs>\t+) {1,3}(?P<rest>.*)",
        lambda match: "{}\t{}".format(match.group("tabs"), match.group("rest")),
        content,
    )


def insert_void(content: str) -> str:
    return re.sub(
        r"(?P<funcdef>[0-9a-zA-Z_]+\t+[0-9a-zA-Z_]*\s*)\(\s*\)",
        lambda match: "{}({})".format(match.group("funcdef"), "void"),
        content,
    )



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    align.py                                           :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/04 09:56:31 by cacharle          #+#    #+#              #
#    Updated: 2021/02/11 20:13:11 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #


import re
from enum import Enum

from c_formatter_42.formatters import helper


class Scope(Enum):
    LOCAL  = 0
    GLOBAL = 1


def align_scope(content: str, scope: Scope) -> str:
    """ Align content
        scope can be either local or global
          local:  for variable declarations in function
          global: for function prototypes
    """

    lines = content.split("\n")
    aligned = []
    # select regex according to scope
    if scope is Scope.LOCAL:
        align_regex = (
            "^\t"
            r"(?P<prefix>{type})\s+"
            r"(?P<suffix>\**{decl};)$"
        )
    elif scope is Scope.GLOBAL:
        align_regex = (
            r"^(?P<prefix>{type})\s+"
            r"(?P<suffix>({name}\(.*\)?;?)|({decl}(;|(\s+=\s+.*))))$"
        )
    align_regex = align_regex.format(
        type=helper.REGEX_TYPE,
        name=helper.REGEX_NAME,
        decl=helper.REGEX_DECL_NAME
    )
    # get the lines to be aligned
    matches = [re.match(align_regex, line) for line in lines]
    aligned = [(i, match.group("prefix"), match.group("suffix"))
               for i, match in enumerate(matches)
               if match is not None and
               match.group("prefix") not in ["struct", "union", "enum"]]

    # global type declaration (struct/union/enum)
    if scope is Scope.GLOBAL:
        typedecl_open_regex  = (r"^(?P<prefix>\s*(typedef\s+)?(struct|enum|union))"
                                r"\s*(?P<suffix>[a-zA-Z_]\w+)?$")
        typedecl_close_regex = r"^(?P<prefix>\})\s*(?P<suffix>([a-zA-Z_]\w+)?;)$"
        in_type_scope = False
        for i, line in enumerate(lines):
            m = re.match(typedecl_open_regex, line)
            if m is not None:
                in_type_scope = True
                if m.group("suffix") is not None and "typedef" not in m.group("prefix"):
                    aligned.append((i, m.group("prefix"), m.group("suffix")))
                continue
            m = re.match(typedecl_close_regex, line)
            if m is not None:
                in_type_scope = False
                if line != "};":
                    aligned.append((i, m.group("prefix"), m.group("suffix")))
                continue
            if in_type_scope:
                m = re.match(
                    r"^(?P<prefix>\s+{type})\s+"
                    r"(?P<suffix>\**{decl};)$"
                    .format(type=helper.REGEX_TYPE, decl=helper.REGEX_DECL_NAME),
                    line
                )
                if m is not None:
                    aligned.append((i, m.group("prefix"), m.group("suffix")))

    # get the minimum alignment required for each line
    min_alignment = max(
        (len(prefix.replace("\t", " " * 4)) // 4 + 1 for _, prefix, _ in aligned),
        default=1
    )
    for i, prefix, suffix in aligned:
        alignment = len(prefix.replace("\t", " " * 4)) // 4
        lines[i] = prefix + "\t" * (min_alignment - alignment) + suffix
        if scope is Scope.LOCAL:
            lines[i] = "\t" + lines[i]
    return "\n".join(lines)


@helper.locally_scoped
def align_local(content: str) -> str:
    """ Wrapper for align_scope to use local_scope decorator """
    return align_scope(content, scope=Scope.LOCAL)


def align(content: str) -> str:
    """ Align the content in global and local scopes """
    content = align_scope(content, scope=Scope.GLOBAL)
    content = align_local(content)
    return content



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    helper.py                                          :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/04 11:38:00 by cacharle          #+#    #+#              #
#    Updated: 2021/02/08 18:22:06 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import re


# regex for a type
REGEX_TYPE      = r"(?!return)([a-z]+\s+)*[a-zA-Z_]\w*"
# regex for a c variable/function name
REGEX_NAME      = r"\**[a-zA-Z_*()]\w*"
# regex for a name in a declaration context (with array and function ptr)
REGEX_DECL_NAME = r"\(?{name}(\[.*\])*(\)\(.*\))?".format(name=REGEX_NAME)


def locally_scoped(func):
    """ Apply the formatter on every local scopes of the content """

    def wrapper(content: str) -> str:
        return re.sub(
            # `*?` is the non greedy version of `*`
            # https://docs.python.org/3/howto/regex.html#greedy-versus-non-greedy
            r"\)\n\{\n(?P<body>.*?)\n\}\n".replace(r"\n", "\n"),
            lambda match: ")\n{\n" + func(match.group("body")) + "\n}\n",
            content,
            flags=re.DOTALL
        )
    return wrapper



import re
from c_formatter_42.formatters import helper


def line_breaker(content: str, column_limit: int = 80) -> str:
    lines = content.split("\n")
    lines = list(map(lambda s: insert_break(s, column_limit), lines))

    return "\n".join(lines)


def insert_break(line: str, column_limit: int) -> str:
    if line_length(line) <= column_limit:
        return line

    # break at all breakable spaces (space after comma or space before binary
    # operators or logical AND or OR)
    breakable_space_pattern = r"((?<=,) | (?=[+\-*/%]|\|\||&&)(?!\*+\S|\+\+|\-\-))"
    line = re.sub(breakable_space_pattern, "\n", line)
    segments = line.split("\n")

    line_indent_level = indent_level(line)

    # join as many segments as it doesn't exceed line length limit
    line = segments[0]
    current_line_length = line_length(segments[0])
    for segment in segments[1:]:
        current_line_length += line_length(segment) + 1
        if current_line_length > column_limit:
            tabulation = "\t" * (line_indent_level + additional_indent_level(line))
            line = ("\n" + tabulation).join([line, segment])
            current_line_length = line_length(tabulation + segment)
        else:
            line = " ".join([line, segment])

    return line


#
# additional indent level increases in proportion to corresponds paren depth
#
# (examples)
# foo() * bar() * baz()
# ~~~~~~~~~~~~~^~~~~~~~ when line breaks here,
# foo() * bar()
# >   * baz()           next line should be indented with 1 tab
# ===
# (foo(bar() * baz()))
# ~~~~~~~~~~^~~~~~~~~    when line breaks here,
# (foo(bar()
# >   >   >   * baz()))  next line should be indented with 3 tabs
#
# function declaration, user defined type (?), and control statement needs one less tab (discount)
#
def additional_indent_level(s: str) -> int:
    additional_indent_level = 1

    discount_pattern = r"(^\t*{type}\t+.*?[a-zA-Z0-9_]\((?!.*?;))|(^\t*typedef)" \
        "|(^\t*(if|while|return))"
    discount_pattern = discount_pattern.format(
        type=helper.REGEX_TYPE,
    )
    if re.match(discount_pattern, s):
        additional_indent_level = 0

    is_surrounded_sq = False
    is_surrounded_dq = False

    for c in s:
        if c == "'":
            is_surrounded_sq = not is_surrounded_sq
        elif c == '"':
            is_surrounded_dq = not is_surrounded_dq
        elif c == "(" and not is_surrounded_sq and not is_surrounded_dq:
            additional_indent_level += 1
        elif c == ")" and not is_surrounded_sq and not is_surrounded_dq:
            additional_indent_level -= 1

    return additional_indent_level


def line_length(line: str) -> int:
    line = line.expandtabs(4)
    return len(line)


def indent_level(line: str) -> int:
    # an exeptional rule for function declaration
    align_pattern = r"^(static\s+)?{type}\s+{name}\((.|\s)*?\);"
    align_pattern = align_pattern.format(
        type=helper.REGEX_TYPE,
        name=helper.REGEX_NAME
    )
    if re.match(align_pattern, line):
        last_tab_index = line.rfind("\t")
        if last_tab_index == -1:
            return 0
        return int(len(line[:last_tab_index + 1].expandtabs(4)) / 4)

    return line.count("\t")



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    clang_format.py                                    :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/04 10:40:07 by cacharle          #+#    #+#              #
#    Updated: 2021/02/25 20:46:18 by cacharle         ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import os
import sys
import inspect
import subprocess
from contextlib import contextmanager

import c_formatter_42.data

CONFIG_FILENAME = ".clang-format"

DATA_DIR = os.path.dirname(inspect.getfile(c_formatter_42.data))


@contextmanager
def _config_context():
    """ Temporarly place .clang-format config file in the current directory
        If there already is a config in the current directory, it's backed up
        then put back in place after clang-format is done running
    """
    config_path = os.path.join(DATA_DIR, CONFIG_FILENAME)
    previous_config = None
    try:
        os.symlink(config_path, CONFIG_FILENAME)
    except FileExistsError:
        if not os.path.islink(CONFIG_FILENAME):
            with open(CONFIG_FILENAME) as f:
                previous_config = f.read()
        os.unlink(CONFIG_FILENAME)
        os.symlink(config_path, CONFIG_FILENAME)
    try:
        yield
    finally:
        os.unlink(CONFIG_FILENAME)
        if previous_config is not None:
            with open(CONFIG_FILENAME, "w") as f:
                f.write(previous_config)


if sys.platform == "linux":
    CLANG_FORMAT_EXEC = os.path.join(DATA_DIR, "clang-format-linux")
elif sys.platform == "darwin":
    CLANG_FORMAT_EXEC = os.path.join(DATA_DIR, "clang-format-darwin")
else:
    raise NotImplementedError("Your platform is not supported")


def clang_format(content: str) -> str:
    """ Wrapper around clang-format

        Pass content on stdin and return stdout.
        The clang-format executable is selected according to the platform,
        this is to keep the same version of clang-format accross all setup.
    """
    with _config_context():
        process = subprocess.Popen(
            [CLANG_FORMAT_EXEC, "-style=file"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = process.communicate(input=content.encode())
        if process.returncode != 0:
            raise RuntimeError(f"clang-format error: {err.decode()}")
        return out.decode()
